# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import Enum
from typing import Sequence

__all__: Sequence[str] = (
    "CodeBlockLang",
    "BlockColor",
    "FunctionFormat",
    "NumberFormats",
    "PropertyColor",
)


class CodeBlockLang(str, Enum):
    abap = "abap"
    arduino = "arduino"
    bash = "bash"
    basic = "basic"
    c = "c"
    clojure = "clojure"
    coffeescript = "coffeescript"
    cplusplus = "c++"
    csharp = "c#"
    css = "css"
    dart = "dart"
    diff = "diff"
    docker = "docker"
    elixir = "elixir"
    elm = "elm"
    erlang = "erlang"
    flow = "flow"
    fortran = "fortran"
    fsharp = "f#"
    gherkin = "gherkin"
    glsl = "glsl"
    go = "go"
    graphql = "graphql"
    groovy = "groovy"
    haskell = "haskell"
    html = "html"
    java = "java"
    javascript = "javascript"
    json = "json"
    julia = "julia"
    kotlin = "kotlin"
    latex = "latex"
    less = "less"
    lisp = "lisp"
    livescript = "livescript"
    lua = "lua"
    makefile = "makefile"
    markdown = "markdown"
    markup = "markup"
    matlab = "matlab"
    mermaid = "mermaid"
    nix = "nix"
    objective_c = "objective-c"
    ocaml = "ocaml"
    pascal = "pascal"
    perl = "perl"
    php = "php"
    plain_text = "plain text"
    powershell = "powershell"
    prolog = "prolog"
    protobuf = "protobuf"
    python = "python"
    r = "r"
    reason = "reason"
    ruby = "ruby"
    rust = "rust"
    sass = "sass"
    scala = "scala"
    scheme = "scheme"
    scss = "scss"
    shell = "shell"
    sql = "sql"
    swift = "swift"
    typescript = "typescript"
    vb_net = "vb.net"
    verilog = "verilog"
    vhdl = "vhdl"
    visual_basic = "visual basic"
    webassembly = "webassembly"
    xml = "xml"
    yaml = "yaml"
    java_or_c = "java/c/c++/c#"


class FunctionFormat(str, Enum):
    average = "average"
    checked = "checked"
    count_ = "count"
    count_per_group = "count_per_group"
    count_values = "count_values"
    date_range = "date_range"
    earliest_date = "earliest_date"
    empty = "empty"
    latest_date = "latest_date"
    max = "max"
    median = "median"
    min = "min"
    not_empty = "not_empty"
    percent_checked = "percent_checked"
    percent_empty = "percent_empty"
    percent_not_empty = "percent_not_empty"
    percent_per_group = "percent_per_group"
    percent_unchecked = "percent_unchecked"
    range = "range"
    show_original = "show_original"
    show_unique = "show_unique"
    sum = "sum"
    unchecked = "unchecked"
    unique = "unique"


class NumberFormats(str, Enum):
    number = "number"
    number_with_commas = "number_with_commas"
    percent = "percent"
    dollar = "dollar"
    canadian_dollar = "canadian_dollar"
    euro = "euro"
    pound = "pound"
    yen = "yen"
    ruble = "ruble"
    rupee = "rupee"
    won = "won"
    yuan = "yuan"
    real = "real"
    lira = "lira"
    rupiah = "rupiah"
    franc = "franc"
    hong_kong_dollar = "hong_kong_dollar"
    new_zealand_dollar = "new_zealand_dollar"
    krona = "krona"
    norwegian_krone = "norwegian_krone"
    mexican_peso = "mexican_peso"
    rand = "rand"
    new_taiwan_dollar = "new_taiwan_dollar"
    danish_krone = "danish_krone"
    zloty = "zloty"
    baht = "baht"
    forint = "forint"
    koruna = "koruna"
    shekel = "shekel"
    chilean_peso = "chilean_peso"
    philippine_peso = "philippine_peso"
    dirham = "dirham"
    colombian_peso = "colombian_peso"
    riyal = "riyal"
    ringgit = "ringgit"
    leu = "leu"
    argentine_peso = "argentine_peso"
    uruguayan_peso = "uruguayan_peso"
    singapore_dollar = "singapore_dollar"


class BlockColor(str, Enum):
    default = "default"
    gray = "gray"
    brown = "brown"
    orange = "orange"
    yellow = "yellow"
    green = "green"
    blue = "blue"
    purple = "purple"
    pink = "pink"
    red = "red"
    gray_background = "gray_background"
    brown_background = "brown_background"
    orange_background = "orange_background"
    yellow_background = "yellow_background"
    green_background = "green_background"
    blue_background = "blue_background"
    purple_background = "purple_background"
    pink_background = "pink_background"
    red_background = "red_background"


class PropertyColor(str, Enum):
    """Color options for database property objects: select/multi_select/status."""

    default = "default"
    gray = "gray"
    brown = "brown"
    orange = "orange"
    red = "red"
    green = "green"
    blue = "blue"
    purple = "purple"
    pink = "pink"
    yellow = "yellow"



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Rich text objects contain the data that Notion uses to display 
formatted text,mentions, and inline equations. 
Arrays of rich text objects within database property objects and page property value objects 
are used to create what a user experiences as a single text value in Notion.
"""
from __future__ import annotations

from typing import Any, Mapping, Optional, Sequence, Union

from notion.properties.build import NotionObject
from notion.properties.common import NotionURL, UserObject
from notion.properties.options import BlockColor

__all__: Sequence[str] = (
    "Annotations",
    "RichText",
    "Equation",
    "Mention",
)


class RichText(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        content: Optional[str] = None,
        /,
        annotations: Optional[Annotations] = None,
        *,
        link: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.set("type", "text")
        self.nest("text", "content", content)
        self.nest("text", "link", NotionURL(link)) if link else None
        if annotations is not None and annotations != {}:
            self.set("annotations", annotations)


class Equation(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self, expression: str, /, *, annotations: Optional[Annotations] = None
    ) -> None:
        super().__init__()
        self.set("type", "equation")
        self.nest("equation", "expression", expression)
        if annotations is not None and annotations != {}:
            self.set("annotations", annotations)


class Mention(NotionObject):
    """
    Classmethods: `user`, `today`, `database`, `page`, `link_preview`, `date`.
    """

    __slots__: Sequence[str] = ()

    def __init__(
        self,
        /,
        *,
        type: str,
        mention_type_object: Union[UserObject, NotionObject],
        annotations: Optional[Annotations] = None,
    ) -> None:
        super().__init__()
        self.set("type", "mention")
        self.nest("mention", "type", type)
        self.nest("mention", type, mention_type_object)

        if annotations is not None and annotations != {}:
            self.set("annotations", annotations)

    @classmethod
    def user(
        cls,
        mention_type_object: Union[UserObject, NotionObject],
        /,
        *,
        annotations: Optional[Annotations] = None,
    ) -> Mention:
        """
        Cannot mention Bots
        :raises: `notion.exceptions.errors.NotionValidationError` Content creation Failed.

        https://developers.notion.com/reference/rich-text#user-mention-type-object
        """
        return cls(
            type="user",
            mention_type_object=mention_type_object,
            annotations=annotations,
        )

    # @classmethod
    # def now(cls, *, annotations: Optional[Annotations] = None) -> Mention:
    #     """https://developers.notion.com/reference/rich-text#template-mention-type-object"""
    #     _mention_type_object = NotionObject()
    #     _mention_type_object.set("type", "template_mention_date")
    #     _mention_type_object.set("template_mention_date", "now")
    #     return cls(
    #         type="template_mention",
    #         mention_type_object=_mention_type_object,
    #         annotations=annotations,
    #     )

    @classmethod
    def today(cls, *, annotations: Optional[Annotations] = None) -> Mention:
        """https://developers.notion.com/reference/rich-text#template-mention-type-object"""
        _mention_type_object = NotionObject()
        _mention_type_object.set("type", "template_mention_date")
        _mention_type_object.set("template_mention_date", "today")
        return cls(
            type="template_mention",
            mention_type_object=_mention_type_object,
            annotations=annotations,
        )

    @classmethod
    def database(
        cls, database_id: str, *, annotations: Optional[Annotations] = None
    ) -> Mention:
        """https://developers.notion.com/reference/rich-text#database-mention-type-object"""
        _mention_type_object = NotionObject()
        _mention_type_object.set("id", database_id)
        return cls(
            type="database",
            mention_type_object=_mention_type_object,
            annotations=annotations,
        )

    @classmethod
    def page(
        cls, page_id: str, *, annotations: Optional[Annotations] = None
    ) -> Mention:
        """https://developers.notion.com/reference/rich-text#page-mention-type-object"""
        _mention_type_object = NotionObject()
        _mention_type_object.set("id", page_id)
        return cls(
            type="page",
            mention_type_object=_mention_type_object,
            annotations=annotations,
        )

    @classmethod
    def link_preview(
        cls, url: str, *, annotations: Optional[Annotations] = None
    ) -> Mention:
        """https://developers.notion.com/reference/rich-text#link-preview-mention-type-object"""
        _mention_type_object = NotionObject()
        _mention_type_object.set("url", url)
        return cls(
            type="link_preview",
            mention_type_object=_mention_type_object,
            annotations=annotations,
        )

    @classmethod
    def date(
        cls,
        start: str,
        end: Optional[str] = None,
        *,
        annotations: Optional[Annotations] = None,
    ) -> Mention:
        """https://developers.notion.com/reference/rich-text#link-preview-mention-type-object"""
        _mention_type_object = NotionObject()
        _mention_type_object.set("start", start)
        _mention_type_object.set("end", end) if end else None
        return cls(
            type="date",
            mention_type_object=_mention_type_object,
            annotations=annotations,
        )


class Annotations(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        bold: Optional[bool] = None,
        italic: Optional[bool] = None,
        strike: Optional[bool] = None,
        underline: Optional[bool] = None,
        code: Optional[bool] = None,
        color: Optional[BlockColor] = None,
    ) -> None:
        super().__init__()

        self.set("bold", bold) if bold else None
        self.set("italic", italic) if italic else None
        self.set("strike", strike) if strike else None
        self.set("underline", underline) if underline else None
        self.set("color", color) if color else None
        self.set("color", BlockColor.default) if not color else None
        self.set("code", code) if code else None

        if not any([[bold, italic, strike, underline, code, color]]):
            pass  # annotations must be defined or non-existent, or Notion will return error



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Sequence, Union

from notion.properties.build import NotionObject
from notion.properties.propertyobjects import PropertyObject
from notion.properties.propertyvalues import PagePropertyValue, TitlePropertyValue

__all__: Sequence[str] = ["Properties"]


class Properties(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, *properties: Union[PropertyObject, PagePropertyValue]) -> None:
        super().__init__()

        for prop in properties:
            if not hasattr(prop, "name"):
                raise AttributeError(
                    (
                        "`notion.properties.propbuild.Properties` ",
                        "is only used for combining named properties. ",
                        "Check to see if `property_name` has been assigned. ",
                    )
                )

            if isinstance(prop, TitlePropertyValue):
                self.nest("properties", prop.name, prop.get("title"))
            else:
                self.nest("properties", prop.name, prop)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from datetime import datetime
from functools import cached_property, reduce
from operator import getitem
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    MutableMapping,
    Optional,
    Sequence,
    Union,
)

from jsonpath_ng.ext import parse  # type: ignore[import]

from notion.api.blockmixin import _TokenBlockMixin
from notion.api.client import _NLOG
from notion.api.notionblock import Block
from notion.api.notiondatabase import Database
from notion.exceptions.errors import NotionInvalidJson, NotionObjectNotFound
from notion.properties import *
from notion.properties.build import build_payload

if TYPE_CHECKING:
    from datetime import timedelta

__all__: Sequence[str] = ["Page"]


class Page(_TokenBlockMixin):
    """
    The Page object contains the page property values of a single Notion page.
    All pages have a Parent. If the parent is a database,
    the property values conform to the schema laid out database's properties.
    Otherwise, the only property value is the title.

    Page content is available as blocks.
    Content can be read using retrieve block children and appended using append block children.

    NOTE: The current version of the Notion api does not allow pages to be created
    to the parent `workspace`. This is why it's necessary to pass a parent instance
    `notion.api.notionpage.Page` or `notion.api.notiondatabase.Database` to the `create`
    class methods, to provide a valid parent id.

    ---
    :param id: (required) `page_id` of object in Notion.
    :param token: (required) Bearer token provided when you create an integration.
        set as `NOTION_TOKEN` in .env or set variable here.
        see https://developers.notion.com/reference/authentication.
    :param notion_version: (optional) API version
        see https://developers.notion.com/reference/versioning

    https://developers.notion.com/reference/page
    """

    def __init__(
        self,
        id: str,
        /,
        *,
        token: Optional[str] = None,
        notion_version: Optional[str] = None,
    ) -> None:
        super().__init__(id, token=token, notion_version=notion_version)

        self.logger = _NLOG.getChild(f"{self.__repr__()}")

    @classmethod
    def create(
        cls, parent_instance: Union[Page, Database, Block], /, *, page_title: str
    ) -> Page:
        """
        Creates a blank page with properties.
        Follow with class methods to add values to properties described in parent database schema,
        or append block children to include content in the page.

        ---
        :param parent_instance: (required) an instance of
            `notion.api.notionpage.Page` or `notion.api.notiondatabase.Database`.
        :param page_title: (required)
        :param icon_url: (optional) #not yet implemented
        :param cover: (optional) #not yet implemented

        https://developers.notion.com/reference/post-page
        """
        if parent_instance.type == "child_database":
            payload = build_payload(
                Parent.database(parent_instance.id),
                Properties(TitlePropertyValue([RichText(page_title)])),
            )
        else:
            payload = build_payload(
                Parent.page(parent_instance.id),
                Properties(TitlePropertyValue([RichText(page_title)])),
            )

        new_page = cls._post(parent_instance, cls._pages_endpoint(), payload=payload)

        cls_ = cls(new_page["id"])
        cls_.logger.info(f"Page created in {parent_instance.__repr__()}")
        cls_.logger.info(f"Url: {new_page['url']}")

        return cls_

    def __getitem__(self, property_name: str) -> MutableMapping[str, Any]:
        try:
            return self.properties[property_name]  # type: ignore[no-any-return]
        except KeyError:
            raise NotionObjectNotFound(
                f"{property_name} not found in page property values."
            )

    @cached_property
    def _retrieve(self) -> MutableMapping[str, Any]:
        return self.retrieve(filter_properties=None)

    @cached_property
    def properties(self) -> MutableMapping[str, Any]:
        return self._retrieve["properties"]  # type: ignore[no-any-return]

    @property
    def title(self) -> str:
        title_keys = ["properties", "title", "title", 0, "text", "content"]
        try:
            return str(reduce(getitem, title_keys, self._retrieve))  # type: ignore[arg-type]
        except IndexError:
            return ""  # page title is empty.

    @title.setter
    def title(self, __new_title: str) -> None:
        self._patch_properties(
            payload=Properties(TitlePropertyValue([RichText(__new_title)]))
        )

    @property
    def url(self) -> MutableMapping[str, Any]:
        return self._retrieve["url"]  # type: ignore[no-any-return]

    @property
    def icon(self) -> MutableMapping[str, Any]:
        return self._retrieve["icon"]  # type: ignore[no-any-return]

    @property
    def cover(self) -> MutableMapping[str, Any]:
        return self._retrieve["cover"]  # type: ignore[no-any-return]

    @property
    def delete_self(self) -> None:
        self._delete(self._block_endpoint(self.id))
        self.logger.info("Deleted self.")

    @property
    def restore_self(self) -> None:
        self._patch(self._pages_endpoint(self.id), payload=(b'{"archived": false}'))
        self.logger.info("Restored self.")

    def retrieve(
        self, *, filter_properties: Optional[list[str]] = None
    ) -> MutableMapping[str, Any]:
        """
        Retrieves a Page object using the ID specified.

        ---
        :param filter_properties: (optional) A list of page property value IDs associated with the page.
            Use this param to limit the response to a specific page property value or values.
            To retrieve multiple properties, specify each page property ID.
            E.g. ?filter_properties=iAk8&filter_properties=b7dh.

        https://developers.notion.com/reference/retrieve-a-page
        """
        if filter_properties:
            _pages_endpoint_filtered_prop = f"{self._pages_endpoint(self.id)}?"
            for name in filter_properties:
                name_id = self.properties[name]["id"]
                _pages_endpoint_filtered_prop += f"filter_properties={name_id}&"
            return self._get(_pages_endpoint_filtered_prop)
        else:
            return self._get(self._pages_endpoint(self.id))

    def _retrieve_property_id(self, property_name: str) -> str:
        """Internal function to retrieve the id of a property.

        :raises: `notion.exceptions.errors.NotionInvalidJson`
        """
        if property_name in self.properties:
            return str(self.properties[property_name]["id"])
        else:
            raise NotionInvalidJson("Property name not found in parent schema.")

    def retrieve_property_item(
        self,
        property_name: str,
    ) -> MutableMapping[str, Any]:
        """
        Retrieves a property_item object for a given page_id and property_id.
        The object returned will either be:
            - a value.
            - a paginated list of property item values.

        ---
        :param property_name: (required) property name in Notion *case-sensitive
            this endpoint only works with property_id's, internal function will retrieve this.
        :param results_only: if true, returns the `results` key index[0] for paginated responses.
            will be either a single dictionary or a list of dictionaries.
        :param property_item_only: if true, returns the `property_item` key.

        https://developers.notion.com/reference/retrieve-a-page-property
        """
        return self._get(
            self._pages_endpoint(
                self.id,
                properties=True,
                property_id=self._retrieve_property_id(property_name),
            )
        )

    def _patch_properties(
        self,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:
        """
        Updates page property values for the specified page.
        Properties not set via the properties parameter will remain unchanged.
        If the parent is a database,
        new property values must conform to the parent database's property schema.

        https://developers.notion.com/reference/patch-page
        """
        return self._patch(self._pages_endpoint(self.id), payload=payload)

    def retrieve_page_content(
        self,
        start_cursor: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> MutableMapping[str, Any]:
        """
        Returns only the first level of children for the specified block.
        See block objects for more detail on determining if that block has nested children.
        In order to receive a complete representation of a block, you
        may need to recursively retrieve block children of child blocks.
        page_size Default: 100 page_size Maximum: 100.

        https://developers.notion.com/reference/get-block-children
        """
        return Block(self.id).retrieve_children(
            page_size=page_size, start_cursor=start_cursor
        )

    def _append(
        self,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:
        """
        Used internally by `notion.api.blocktypefactory.BlockFactory`.

        https://developers.notion.com/reference/patch-block-children
        """
        return self._patch(
            self._block_endpoint(self.id, children=True), payload=payload
        )

    def set_select(self, column_name: str, select_option: str) -> None:
        """
        :param select_option: (required) if the option already exists, then it is
            case sensitive. if the option does not exist, it will be created.
        """
        color = [
            m.value
            for m in parse(
                f"$.select.options[?(@.name=='{select_option}')].color"
            ).find(Database(self.parent_id)[column_name])
        ]

        if color[0]:
            self._patch_properties(
                Properties(
                    SelectPropertyValue(column_name, Option(select_option, color[0]))
                )
            )
        else:
            self._patch_properties(
                Properties(SelectPropertyValue(column_name, Option(select_option)))
            )

    def set_multiselect(
        self, column_name: str, multi_select_options: list[str]
    ) -> None:
        """
        :param multi_select_options: (required) list of strings for each select option.
            if the option already exists, then it is case sensitive.
            if the option does not exist, it will be created.
        """
        selected_options: list[Option] = []

        for option in multi_select_options:
            color = [
                m.value
                for m in parse(
                    f"$.multi_select.options[?(@.name=='{option}')].color"
                ).find(Database(self.parent_id)[column_name])
            ]
            if color[0]:
                selected_options.append(Option(option, color[0]))
            else:
                selected_options.append(Option(option))

        self._patch_properties(
            Properties(MultiSelectPropertyValue(column_name, selected_options))
        )

    def set_status(self, column_name: str, status_option: str) -> None:
        """
        :param status_option: (required) unlike select/multi-select,
            status option must already exist when using this endpoint.
            to create a new status option, use the database endpoints.
        """
        color = [
            m.value
            for m in parse(
                f"$.status.options[?(@.name=='{status_option}')].color"
            ).find(Database(self.parent_id)[column_name])
        ]

        if color[0]:
            self._patch_properties(
                Properties(
                    StatusPropertyValue(column_name, Option(status_option, color[0]))
                )
            )
        else:
            self._patch_properties(
                Properties(StatusPropertyValue(column_name, Option(status_option)))
            )

    def set_date(
        self,
        column_name: str,
        start: Union[str, datetime],
        end: Optional[Union[str, datetime]] = None,
    ) -> None:
        """
        :param start: (required) A date, with an optional time.
            If the "date" value is a range, then start represents the start of the range.
        :param end: (optional) A string representing the end of a date range.
            If the value is null, then the date value is not a range.
        """
        if isinstance(start, datetime):
            start = start.replace().astimezone(self.tz)
        if end and isinstance(end, datetime):
            end = end.replace().astimezone(self.tz)

        self._patch_properties(
            Properties(DatePropertyValue(column_name, start=start, end=end))
        )

    def set_text(self, column_name: str, new_text: Union[str, Any]) -> None:
        """
        :param new_text: (required) to replace the current text
        """
        self._patch_properties(
            Properties(RichTextPropertyValue(column_name, [RichText(new_text)]))
        )

    def set_files(
        self, column_name: str, array_of_files: list[Union[InternalFile, ExternalFile]]
    ) -> None:
        self._patch_properties(
            Properties(FilesPropertyValue(column_name, array_of_files))
        )

    def set_phonenumber(self, column_name: str, phone_number: str) -> None:
        self._patch_properties(
            Properties(PhoneNumberPropertyValue(column_name, phone_number))
        )

    def set_related(self, column_name: str, related_ids: list[str]) -> None:
        """
        :param related_ids: (required) list of notion page ids to reference
        """
        list_ids: list[NotionUUID] = []
        for id in related_ids:
            list_ids.append(NotionUUID(id))

        self._patch_properties(Properties(RelationPropertyValue(column_name, list_ids)))

    def set_checkbox(self, column_name: str, value: bool) -> None:
        """
        :param value: (required) to replace the current bool
        """
        self._patch_properties(Properties(CheckboxPropertyValue(column_name, value)))

    def set_number(self, column_name: str, new_number: Union[float, timedelta]) -> None:
        """
        :param new_number: (required) to replace the current number
        """
        self._patch_properties(Properties(NumberPropertyValue(column_name, new_number)))

    def set_people(self, column_name: str, user_array: list[UserObject]) -> None:
        self._patch_properties(Properties(PeoplePropertyValue(column_name, user_array)))

    def set_email(self, column_name: str, email: str) -> None:
        self._patch_properties(Properties(EmailPropertyValue(column_name, email)))

    def set_url(self, column_name: str, url: str) -> None:
        self._patch_properties(Properties(URLPropertyValue(column_name, url)))



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, MutableMapping, Sequence, Union

from notion.exceptions.errors import *

__all__: Sequence[str] = ["validate_response"]


def validate_response(response: MutableMapping[str, Any]) -> Union[_NotionErrors, None]:
    r"""
    To be used on responses from notion.Page, notion.Database, notion.Block.

    Example:
    ```py
    page = notion.Page("12345")

    # Object for invalid Notion requests:
    # {
    #   'object': 'error',
    #   'status': 400,
    #   'code': 'validation_error',
    #   'message': 'path failed validation: path.page_id should be a valid uuid, instead was `"12345"`.'
    # }
    ```
    ------
    ```sh
    Traceback (most recent call last):
      File "c:\path\to\file\_.py", line 212, in <module>
        validate_response(response)
      File "c:\path\to\file\_.py", line 186, in validate_response
        raise NotionValidationError(message)
    notion.exceptions.errors.NotionValidationError: path failed validation: path.page_id should be a valid uuid, instead was `"12345"`.
    Error 400:
    The request body does not match the schema for the expected parameters.
    ```

    https://developers.notion.com/reference/errors
    """
    if "error" in response.values():
        code = response["code"]
        message = response["message"]

        if "invalid_json" in code:
            raise NotionInvalidJson(message)
        if "invalid_request_url" in code:
            raise NotionInvalidRequestUrl(message)
        if "invalid_request" in code:
            raise NotionInvalidRequest(message)
        if "validation_error" in code:
            raise NotionValidationError(message)
        if "missing_version" in code:
            raise NotionMissingVersion(message)
        if "unauthorized" in code:
            raise NotionUnauthorized(message)
        if "restricted_resource" in code:
            raise NotionRestrictedResource(message)
        if "object_not_found" in code:
            raise NotionObjectNotFound(message)
        if "conflict_error" in code:
            raise NotionConflictError(message)
        if "rate_limited" in code:
            raise NotionRateLimited(message)
        if "internal_server_error" in code:
            raise NotionInternalServerError(message)
        if "service_unavailable" in code:
            raise NotionServiceUnavailable(message)
        if "database_connection_unavailable" in code:
            raise NotionDatabaseConnectionUnavailable(message)

    return None



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from functools import cached_property
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    MutableMapping,
    Optional,
    Sequence,
    Union,
)

try:
    import orjson

    default_json: ModuleType = orjson
except ModuleNotFoundError:
    import json

    default_json: ModuleType = json  # type: ignore[no-redef]

from notion.api.blockmixin import _TokenBlockMixin
from notion.api.client import _NLOG
from notion.exceptions.errors import NotionInvalidRequest, NotionObjectNotFound
from notion.properties import *

if TYPE_CHECKING:
    from notion.api.notionpage import Page

__all__: Sequence[str] = ["Database"]


class Database(_TokenBlockMixin):
    """
    Database objects describe the property schema of a database in Notion.
    Pages are the items (or children) in a database. The API treats database rows as pages.

    All databases require one, and only one, title property.
    The API throws errors if you send a request to Create a database without a title property,
    or if you attempt to Update a database to add or remove a title property.

    ### Title database property vs. database title
    A title database property is a type of column in a database.
    A database title defines the title of the database and is found on the database object.
    Every database requires both a database title and a title database property.

    NOTE: Adding a new column with the same name as an existing column will replace the existing column
    with the new column type. If the column of the same name is updated again, but back to the original
    column type - the previous settings (number format, formula expression, etc..) will reapply.

    ---
    :param id: (required) `database_id` of object in Notion.
    :param token: (required) Bearer token provided when you create an integration.
        set as `NOTION_TOKEN` in .env or set variable here.
        see https://developers.notion.com/reference/authentication.
    :param notion_version: (optional) API version
        see https://developers.notion.com/reference/versioning

    ---
    :raises `notion.exceptions.errors.NotionInvalidRequest`: if using an id that does not
        reference a database in Notion.

    https://developers.notion.com/reference/database
    """

    def __new__(cls, id: str, /) -> Database:
        target_block = _TokenBlockMixin(id)
        if target_block.type != "child_database":
            raise NotionInvalidRequest(
                f"{target_block.__repr__()} does not reference a Database"
            )
        return super().__new__(cls)

    def __init__(
        self,
        id: str,
        /,
        *,
        token: Optional[str] = None,
        notion_version: Optional[str] = None,
    ) -> None:
        super().__init__(id, token=token, notion_version=notion_version)

        self.logger = _NLOG.getChild(f"{self.__repr__()}")

    @classmethod
    def create(
        cls, parent_instance: Page, /, *, database_title: str, name_column: str
    ) -> Database:
        """
        Creates a non-inline database in the specified parent page,
        with the specified properties schema.
        Currently, Databases cannot be created to the parent workspace.

        ---
        :param parent_instance: (required) an instance of `notion.api.notionpage.Page`.
        :param database_title: (required) title of new database.
        :param name_column: (required) name for main column for page names.

        :raises `notion.exceptions.errors.NotionInvalidRequest`:
            if trying to create a database directly inside another database.

        https://developers.notion.com/reference/create-a-database
        """
        if parent_instance.type == "child_database":
            raise NotionInvalidRequest("Cannot create a database in a database.")

        parent = Parent.page(parent_instance.id)
        title = TitlePropertyValue([RichText(database_title)])
        properties = Properties(TitlePropertyObject(name_column))
        schema = build_payload(parent, title, properties)

        new_db = cls._post(parent_instance, cls._database_endpoint(), payload=schema)

        cls_ = cls(new_db["id"])
        cls_.logger.info(f"New database created in {parent_instance.__repr__()}")
        cls_.logger.info(f"Url: {new_db['url']}")

        return cls_

    def __getitem__(self, property_name: str) -> MutableMapping[str, Any]:
        try:
            return self._property_schema[property_name]  # type: ignore[no-any-return]
        except KeyError:
            raise NotionObjectNotFound(
                f"{property_name} not found in page property values."
            )

    @cached_property
    def retrieve(self) -> MutableMapping[str, Any]:
        """https://developers.notion.com/reference/retrieve-a-database"""
        return self._get(self._database_endpoint(self.id))

    @cached_property
    def _property_schema(self) -> MutableMapping[str, Any]:
        return self.retrieve["properties"]  # type: ignore[no-any-return]

    @property
    def title(self) -> str:
        try:
            return str(self.retrieve["title"][0]["text"]["content"])
        except IndexError:
            return ""

    @title.setter
    def title(self, __new_title: str) -> None:
        self._update(payload=TitlePropertyValue([RichText(__new_title)]))

    @property
    def inline(self) -> bool:
        """
        Has the value true if the database appears in the page as an inline block.
        Otherwise has the value false if the database appears as a child page.

        returns the current inline status.
        """
        return bool(self.retrieve["is_inline"])

    @inline.setter
    def inline(self, __inline_status: bool) -> None:
        self._patch(
            self._database_endpoint(self.id),
            payload=default_json.dumps({"is_inline": __inline_status}),
        )

    @property
    def url(self) -> str:
        return str(self.retrieve["url"])

    @property
    def icon(self) -> MutableMapping[str, Any]:
        return self.retrieve["icon"]  # type: ignore[no-any-return]

    @property
    def delete_self(self) -> None:
        self._delete(self._block_endpoint(self.id))
        self.logger.info("Deleted self.")

    @property
    def restore_self(self) -> None:
        self._patch(self._database_endpoint(self.id), payload=(b'{"archived": false}'))
        self.logger.info("Restored self.")

    def _update(
        self,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:
        """
        Updates an existing database as specified by the parameters.
        Used internally but optionally can update custom payloads.

        ---
        :param payload: (required) json payload for updated properties parameters.

        https://developers.notion.com/reference/update-a-database
        """
        return self._patch(self._database_endpoint(self.id), payload=payload)

    def delete_property(self, name_or_id: str) -> None:
        """
        :param name_or_id: (required) name or id of property described in database schema

        https://developers.notion.com/reference/update-property-schema-object#removing-a-property
        """
        self._update(payload={"properties": {name_or_id: None}})
        self.logger.info(f"Deleted property `{name_or_id}`")

    def rename_property(self, old_name: str, new_name: str) -> None:
        """
        https://developers.notion.com/reference/update-property-schema-object#renaming-a-property
        """
        self._update(
            payload=default_json.dumps({"properties": {old_name: {"name": new_name}}})
        )
        self.logger.info(f"Renamed property `{old_name}` to `{new_name}`")

    def query(
        self,
        *,
        payload: Optional[
            Union[MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]]
        ] = None,
        filter_property_values: Optional[list[str]] = None,
    ) -> MutableMapping[str, Any]:
        """
        Gets a list of Pages contained in the database,
        filtered/ordered to the filter conditions/sort criteria provided in request.
        The response may contain fewer than page_size of results.
        Responses from paginated endpoints contain a `next_cursor` property,
        which can be used in a query payload to continue the list.
        page_size Default: 100 page_size Maximum: 100.

        ---
        :param payload: (optional) filter/sort objects to apply to query.
            filter objects built in `notion.query`
        :param filter_property_values: (optional) list of property names,
            query will only return the selected properties.

        https://developers.notion.com/reference/post-database-query
        """
        query_url = self._database_endpoint(self.id, query=True)

        if filter_property_values:
            query_url = query_url + "?"
            for name in filter_property_values:
                name_id = self._property_schema[name].get("id")
                query_url += "filter_properties=" + name_id + "&"
            return self._post(query_url, payload=payload)
        else:
            return self._post(query_url, payload=payload)

    def dual_relation_column(
        self, property_name: str, database_id: str, synced_property_name: str
    ) -> None:
        """
        :param database_id: (required) The database that the relation property refers to.
            The corresponding linked page values must belong to the database in order to be valid.
        :param synced_property_name: (required) The name of the corresponding property that is
            updated in the related database when this property is changed.
        """
        self._update(
            Properties(
                RelationPropertyObject.dual(
                    property_name, database_id, synced_property_name
                )
            )
        )

        # NOTE: there is an issue with the current API version and `synced_property_name`,
        # Notion UI will default to `Related to {original database name} ({property name})`,
        # regardless of what name is included in the request.
        # TEMP fix to default synced property name
        synced_database = Database(database_id)
        try:
            synced_database[f"Related to {self.title} ({property_name})"]
            Database(database_id).rename_property(
                f"Related to {self.title} ({property_name})", synced_property_name
            )
        except NotionObjectNotFound:
            pass  # if Notion patches this in between versions

        self.logger.info(
            (
                f"Created/Updated dual_relation property `{property_name}` ",
                f" linked to notion.Database('{database_id}').",
            )
        )

    def single_relation_column(self, property_name: str, database_id: str) -> None:
        self._update(
            Properties(RelationPropertyObject.single(property_name, database_id))
        )
        self.logger.info(
            (
                f"Created/Updated dual_relation property `{property_name}` ",
                f" linked to notion.Database('{database_id}').",
            )
        )

    def rollup_column(
        self,
        property_name: str,
        relation_property_name: str,
        rollup_property_name: str,
        function: Union[FunctionFormat, str],
    ) -> None:
        """
        :param property_name: name for the new rollup column
        :param relation_property_name: name of relation column to other database
        :param rollup_property_name: name of column in other database to calculate
        :param function: `notion.properties.options.FunctionFormat` or refer to api reference.
        """
        self._update(
            (
                Properties(
                    RollupPropertyObject(
                        property_name,
                        relation_property_name,
                        rollup_property_name,
                        function,
                    )
                )
            )
        )
        self.logger.info(f"Created/Updated rollup property `{property_name}`.")

    def select_column_column(
        self, property_name: str, /, *, options: list[Option]
    ) -> None:
        """
        If `property_name` is a select column that already exists,
        then the available options will be overwritten with the new list passed to this method.

        You can set the color of an option when first creating it,
        but if `property_name` already existsand a color is assigned in `notion.properties.Option`,
        then `notion.exceptions.errors.NotionValidationError` will raise.

        You can pass an empty list to `options` to clear the available options,
        and then readd them with custom colors.
        """
        self._update(Properties(SelectPropertyObject(property_name, options=options)))
        self.logger.info(f"Created/Updated select property `{property_name}`.")

    def multiselect_column(
        self, property_name: str, /, *, options: list[Option]
    ) -> None:
        """
        If `property_name` is a multi-select column that already exists,
        then the available options will be overwritten with the new list passed to this method.

        You can set the color of an option when first creating it,
        but if `property_name` already exists and a color is assigned in `notion.properties.Option`,
        then `notion.exceptions.errors.NotionValidationError` will raise.

        You can pass an empty list to `options` to clear the available options,
        and then re-add them with custom colors.
        """
        self._update(
            Properties(MultiSelectPropertyObject(property_name, options=options))
        )
        self.logger.info(f"Created/Updated multi-select property `{property_name}`.")

    def number_column(
        self,
        property_name: str,
        /,
        *,
        format: Optional[Union[NumberFormats, str]] = None,
    ) -> None:
        if not format:
            format = NumberFormats.number
        self._update(Properties(NumberPropertyObject(property_name, format)))
        self.logger.info(f"Created/Updated number property `{property_name}`.")

    def checkbox_column(self, property_name: str) -> None:
        self._update(Properties(CheckboxPropertyObject(property_name)))
        self.logger.info(f"Created/Updated checkbox property `{property_name}`.")

    def date_column(self, property_name: str) -> None:
        self._update(Properties(DatePropertyObject(property_name)))
        self.logger.info(f"Created/Updated date property `{property_name}`.")

    def text_column(self, property_name: str) -> None:
        self._update(Properties(RichTextPropertyObject(property_name)))
        self.logger.info(f"Created/Updated text property `{property_name}`.")

    def formula_column(self, property_name: str, /, *, expression: str) -> None:
        self._update(Properties(FormulaPropertyObject(property_name, expression)))
        self.logger.info(f"Created/Updated formula property `{property_name}`.")

    def created_time_column(self, property_name: str) -> None:
        self._update(Properties(CreatedTimePropertyObject(property_name)))
        self.logger.info(f"Created/Updated created_time property `{property_name}`.")

    def created_by_column(self, property_name: str) -> None:
        self._update(Properties(CreatedByPropertyObject(property_name)))
        self.logger.info(f"Created/Updated created_by property `{property_name}`.")

    def last_edited_time_column(self, property_name: str) -> None:
        self._update(Properties(LastEditedTimePropertyObject(property_name)))
        self.logger.info(f"Created/Updated last_edited property `{property_name}`.")

    def last_edited_by_column(self, property_name: str) -> None:
        self._update(Properties(LastEditedByPropertyObject(property_name)))
        self.logger.info(f"Created/Updated last_edited_by property `{property_name}`.")

    def files_column(self, property_name: str) -> None:
        self._update(Properties(FilesPropertyObject(property_name)))
        self.logger.info(f"Created/Updated files property `{property_name}`.")

    def email_column(self, property_name: str) -> None:
        self._update(Properties(EmailPropertyObject(property_name)))
        self.logger.info(f"Created/Updated email property `{property_name}`.")

    def url_column(self, property_name: str) -> None:
        self._update(Properties(URLPropertyObject(property_name)))
        self.logger.info(f"Created/Updated url property `{property_name}`.")

    def phonenumber_column(self, property_name: str) -> None:
        self._update(Properties(PhoneNumberPropertyObject(property_name)))
        self.logger.info(f"Created/Updated phone_number property `{property_name}`.")

    def people_column(self, property_name: str) -> None:
        self._update(Properties(PeoplePropertyObject(property_name)))
        self.logger.info(f"Created/Updated people property `{property_name}`.")

    # NOTE:
    # It is not possible to update a status database property in the current API version.
    # Update these values from the Notion UI, instead.



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Sequence, Union

from notion.properties.build import NotionObject
from notion.query.propfilter import PropertyFilter
from notion.query.timestamp import TimestampFilter

__all__: Sequence[str] = ["CompoundFilter"]


class CompoundFilter(NotionObject):
    """NOTE: only up to two nesting levels deep.

    :method _and(): combine all filters in an `and` grouping.
    :method _or(): combine all filters in an `or` grouping.

    Create a separate CompoundFilter object to nest an `and` operator inside another `and` or `or`.

    https://developers.notion.com/reference/post-database-query-filter#compound-filter-object
    """

    __slots__: Sequence[str] = ()

    def __init__(self) -> None:
        super().__init__()

    def _and(self, *filters: FilterTypeObjects) -> CompoundFilter:
        filters_ = [f["filter"] if "filter" in f else f for f in filters]
        self.nest("filter", "and", filters_)
        return self

    def _or(self, *filters: FilterTypeObjects) -> CompoundFilter:
        filters_ = [f["filter"] if "filter" in f else f for f in filters]
        self.nest("filter", "or", filters_)
        return self


FilterTypeObjects = Union[PropertyFilter, CompoundFilter, TimestampFilter]



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Sequence, Union

from notion.properties.build import NotionObject
from notion.properties.common import NotionURL
from notion.properties.files import ExternalFile
from notion.properties.options import BlockColor, CodeBlockLang
from notion.properties.richtext import Equation, Mention, RichText

# See docs in `notion.api.blockwrite.BlockFactory` for info on block types.

__all__: Sequence[str] = (
    "Children",
    "OriginalSyncedBlockType",
    "ReferenceSyncedBlockType",
    "ParagraphBlocktype",
    "NewLineBreak",
    "CalloutBlocktype",
    "QuoteBlocktype",
    "BulletedListItemBlocktype",
    "NumberedListItemBlocktype",
    "ToDoBlocktype",
    "ToggleBlocktype",
    "CodeBlocktype",
    "EmbedBlocktype",
    "BookmarkBlocktype",
    "EquationBlocktype",
    "TableOfContentsBlocktype",
    "DividerBlock",
    "BreadcrumbBlock",
    "Heading1BlockType",
    "Heading2BlockType",
    "Heading3BlockType",
    "LinkToPageBlockType",
)


class Children(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        block_type_objects_array: Optional[list[NotionObject]],
    ) -> None:
        super().__init__()
        if not block_type_objects_array:
            block_type_objects_array = []

        self.set("children", block_type_objects_array)


class OriginalSyncedBlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, children: Optional[list[str]]) -> None:
        super().__init__()
        if not children:
            children = []

        self.set("type", "synced_block")
        self.nest("synced_block", "synced_from", None)
        self.nest("synced_block", "children", children)


class ReferenceSyncedBlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, block_id: str) -> None:
        super().__init__()
        self.set("type", "synced_block")
        self.nest("synced_block", "synced_from", {"type": "block_id"})
        self.nest("synced_block", "synced_from", {"block_id": block_id})


class ParagraphBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "paragraph")
        self.nest("paragraph", "rich_text", rich_text)
        self.nest("paragraph", "color", block_color)


NewLineBreak = ParagraphBlocktype([RichText("")])


class CalloutBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        icon: Optional[str] = None,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "callout")
        self.nest("callout", "rich_text", rich_text)
        self.nest("callout", "color", block_color)
        if icon:
            self.nest("callout", "icon", ExternalFile(icon))


class QuoteBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "quote")
        self.nest("quote", "rich_text", rich_text)
        self.nest("quote", "color", block_color)


class BulletedListItemBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "bulleted_list_item")
        self.nest("bulleted_list_item", "rich_text", rich_text)
        self.nest("bulleted_list_item", "color", block_color)


class NumberedListItemBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "numbered_list_item")
        self.nest("numbered_list_item", "rich_text", rich_text)
        self.nest("numbered_list_item", "color", block_color)


class ToDoBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        checked: Optional[bool] = False,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "to_do")
        self.nest("to_do", "rich_text", rich_text)
        self.nest("to_do", "color", block_color)
        self.nest("to_do", "checked", checked)


class ToggleBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "toggle")
        self.nest("toggle", "rich_text", rich_text)
        self.nest("toggle", "color", block_color)


class CodeBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        language: Optional[Union[CodeBlockLang, str]] = None,
        caption: Optional[Sequence[Union[RichText, Mention, Equation]]] = None,
    ) -> None:
        super().__init__()

        if not language:
            language = CodeBlockLang.plain_text

        self.set("type", "code")
        self.nest("code", "rich_text", rich_text)
        self.nest("code", "language", language)
        self.nest("code", "caption", caption) if caption else None


class EmbedBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, embedded_url: str, /) -> None:
        super().__init__()
        self.set("type", "embed")
        self.set("embed", NotionURL(embedded_url))


class BookmarkBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        bookmark_url: str,
        /,
        *,
        caption: Optional[Sequence[Union[RichText, Mention, Equation]]] = None,
    ) -> None:
        super().__init__()
        self.set("type", "bookmark")
        self.set("bookmark", NotionURL(bookmark_url))
        self.nest("bookmark", "caption", caption) if caption else None


class EquationBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, expression: str) -> None:
        super().__init__()
        self.set("type", "equation")
        self.nest("equation", "expression", expression)


class TableOfContentsBlocktype(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, block_color: Optional[Union[BlockColor, str]] = None) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "table_of_contents")
        self.nest("table_of_contents", "color", block_color)


class Heading1BlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "heading_1")
        self.nest("heading_1", "rich_text", rich_text)
        self.nest("heading_1", "color", block_color)
        self.nest("heading_1", "is_toggleable", is_toggleable)


class Heading2BlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "heading_2")
        self.nest("heading_2", "rich_text", rich_text)
        self.nest("heading_2", "color", block_color)
        self.nest("heading_2", "is_toggleable", is_toggleable)


class Heading3BlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(
        self,
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> None:
        super().__init__()

        if not block_color:
            block_color = BlockColor.default

        self.set("type", "heading_3")
        self.nest("heading_3", "rich_text", rich_text)
        self.nest("heading_3", "color", block_color)
        self.nest("heading_3", "is_toggleable", is_toggleable)


class LinkToPageBlockType(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, page_id: str) -> None:
        super().__init__()

        self.set("type", "link_to_page")
        self.nest("link_to_page", "type", "page_id")
        self.nest("link_to_page", "page_id", page_id)


class BreadcrumbBlock(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self) -> None:
        super().__init__()
        self.set("type", "breadcrumb")
        self.set("breadcrumb", {})


class DividerBlock(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self) -> None:
        super().__init__()
        self.set("type", "divider")
        self.set("divider", {})



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""" 
A block object represents content within Notion. Blocks can be text, lists, media, and more. A page is a type of block, too.
Some blocks have more content nested inside them. Some examples are indented paragraphs, lists, and toggles. 
The nested content is called children, and children are blocks, too.

Block types that support children
Block types which support children are:
    "paragraph", "bulleted_list_item", "numbered_list_item", "toggle", "to_do", "quote", "callout", 
    "synced_block", "template", "column", "child_page", "child_database", and "table". 
    All heading blocks ("heading_1", "heading_2", and "heading_3") support children when the is_toggleable property is true.

NOTE: The link_preview block will only be returned as part of a response. It cannot be created via the API.

https://developers.notion.com/reference/block
"""
from __future__ import annotations

from functools import reduce, singledispatchmethod
from operator import getitem, methodcaller
from typing import Any, MutableMapping, Optional, Sequence, Union

from notion.api.client import _NotionClient
from notion.api.notionblock import Block
from notion.api.notionpage import Page
from notion.properties import *

__all__: Sequence[str] = ["BlockFactory"]


class BlockFactory(_NotionClient):
    """
    Factory to write block types as a child of the target
    `notion.api.notionblock.Block` or `notion.api.notionpage.Page`.
    Returns an instance of `notion.api.notionblock.Block`.

    To recurisvely set children blocks, each method will return the new block object,
    which contains the `block_id` that can be extracted for the `target`
    OR
    nest `Blockfactory` with the outermost method being the last nested block.
    ```py
    notion.BlockFactory.divider(
        notion.BlockFactory.paragraph(
            notion.BlockFactory.paragraph(
                notion.Block('87aadab8ce4b407682197c922e51511f'
                ), ['This is the parent']
            ), ['This is the child']
        )
    )
    ```
    will output in Notion:
    ```md
    This is block 87aadab8ce4b407682197c922e51511f
        This is the parent
            This is the child
                ---
    ```
    NOTE: Nested Children
    For blocks that allow children, we allow up to two levels of nesting in a single request.
    """

    def __init__(self) -> None:
        pass

    @singledispatchmethod
    @staticmethod
    def _append(  # type: ignore[empty-body]
        target: Union[Page, Block, MutableMapping[str, Any]],
        payload: MutableMapping[str, Any],
    ) -> Block:
        ...

    @_append.register
    @staticmethod
    def _(target: Union[Page, Block], payload: MutableMapping[str, Any]) -> Block:
        _append_method = methodcaller("_append", payload=payload)
        new_block = _append_method(target)
        id_new_block = str(reduce(getitem, ["results", 0, "id"], new_block))
        return Block(id_new_block)

    @_append.register
    @staticmethod
    def _(target: dict, payload: MutableMapping[str, Any]) -> Block:  # type: ignore[type-arg]
        """
        If target is not an instance of Page or Block,
        It can also be a block object from a list of results
        ```json
        {
            "object": "list",
            "results": [
                {
                    "object": "block",
                    "id": "9dd8e0d6-8370-4f19-914b-39abc9863c1a",
                    // ...
                }
            ],
            "next_cursor": null,
            "has_more": false,
            "type": "block",
            "block": {}
        }
        ```
        or a block object directly
        ```json
        {
            "object": "block",
            "id": "647eec30-e714-4c5f-908f-73ab2f6e8c99"
            // ...
        }
        ```
        """
        if target["object"] == "list":
            assert "type" in target and target["block"] == {}
            id_target = str(target["results"][0]["id"])
            new_block = Block(id_target)._append(payload)
            id_new_block = str(new_block["results"][0]["id"])

            return Block(id_new_block)
        else:
            assert target["object"] == "block"
            new_block = Block(target["id"])._append(payload)
            id_new_block = str(new_block["results"][0]["id"])

            return Block(id_new_block)

    @staticmethod
    def reference_synced_block(
        target: Union[Page, Block, MutableMapping[str, Any]], block_id: str
    ) -> Block:
        """
        ### Reference Synced Block
        To sync the content of the original synced_block with another synced_block,
        the developer simply needs to refer to that synced_block using the synced_from property.

        Note that only "original" synced blocks can be referenced in the synced_from property.

        ---
        :param block_id: (required) string (UUIDv4). Identifier of an original synced_block

        https://developers.notion.com/reference/block#synced-block-blocks
        """

        reference_synced_block_block = BlockFactory._append(
            target, Children([ReferenceSyncedBlockType(block_id)])
        )
        return reference_synced_block_block

    @staticmethod
    def new_synced_block(target: Union[Page, Block, MutableMapping[str, Any]]) -> Block:
        """
        Similar to the UI, there are two versions of a synced_block --
        the original block that was created first and doesn't yet sync with anything else,
        and the reference blocks that are synced to the original synced block.

        ### Original Synced Block
        To create a synced_block, the developer needs to create an original synced block.
        Developers will be able to identify the original synced_block because it does not
        "sync_from" any other block (synced_from property is set to null).

        Note that all of the blocks available to be synced in another synced_block must be captured in the children property.

        `synced_from` Value is always null to signify that this is an original synced block and that is not referring to another block

        ---
        This method will create a new _empty_ synced block that can be
        appended to other blocks with `BlockWrite.reference_synced_block(...)`.
        return it to get the new block id and use as a target to append new children.

        View `notion.properties.blocktypes.OriginalSyncedBlock`, or Notion API reference
        for more information.

        https://developers.notion.com/reference/block#synced-block-blocks
        """

        new_synced_block_block = BlockFactory._append(
            target, Children([OriginalSyncedBlockType(children=[])])
        )
        return new_synced_block_block

    @staticmethod
    def breadcrumb(target: Union[Page, Block, MutableMapping[str, Any]]) -> Block:
        """https://developers.notion.com/reference/block#breadcrumb-blocks"""
        breadcrumb_block = BlockFactory._append(target, Children([BreadcrumbBlock()]))
        return breadcrumb_block

    @staticmethod
    def divider(target: Union[Page, Block, MutableMapping[str, Any]]) -> Block:
        """https://developers.notion.com/reference/block#divider-blocks"""
        divider_block = BlockFactory._append(target, Children([DividerBlock()]))
        return divider_block

    @staticmethod
    def newline(target: Union[Page, Block, MutableMapping[str, Any]]) -> Block:
        newline_block = BlockFactory._append(target, Children([NewLineBreak]))
        return newline_block

    @staticmethod
    def quote(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#quote-blocks"""
        quote_block = BlockFactory._append(
            target, Children([QuoteBlocktype(rich_text, block_color=block_color)])
        )
        return quote_block

    @staticmethod
    def callout(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        icon: Optional[str] = None,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """
        :param icon: url to external source for icon.

        ---
        https://developers.notion.com/reference/block#callout-blocks
        """
        callout_block = BlockFactory._append(
            target,
            Children([CalloutBlocktype(rich_text, icon=icon, block_color=block_color)]),
        )
        return callout_block

    @staticmethod
    def paragraph(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#paragraph-blocks"""
        paragraph_block = BlockFactory._append(
            target, Children([ParagraphBlocktype(rich_text, block_color=block_color)])
        )
        return paragraph_block

    @staticmethod
    def heading1(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> Block:
        """https://developers.notion.com/reference/block#heading-one-blocks"""
        heading1_block = BlockFactory._append(
            target,
            Children(
                [
                    Heading1BlockType(
                        rich_text, block_color=block_color, is_toggleable=is_toggleable
                    )
                ]
            ),
        )
        return heading1_block

    @staticmethod
    def heading2(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> Block:
        """https://developers.notion.com/reference/block#heading-two-blocks"""
        heading2_block = BlockFactory._append(
            target,
            Children(
                [
                    Heading2BlockType(
                        rich_text, block_color=block_color, is_toggleable=is_toggleable
                    )
                ]
            ),
        )
        return heading2_block

    @staticmethod
    def heading3(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        is_toggleable: Optional[bool] = False,
    ) -> Block:
        """https://developers.notion.com/reference/block#heading-three-blocks"""
        heading3_block = BlockFactory._append(
            target,
            Children(
                [
                    Heading3BlockType(
                        rich_text, block_color=block_color, is_toggleable=is_toggleable
                    )
                ]
            ),
        )
        return heading3_block

    @staticmethod
    def bulleted_list(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#bulleted-list-item-blocks"""
        bulleted_list_block = BlockFactory._append(
            target,
            Children([BulletedListItemBlocktype(rich_text, block_color=block_color)]),
        )
        return bulleted_list_block

    @staticmethod
    def numbered_list(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#numbered-list-item-blocks"""
        numbered_list_block = BlockFactory._append(
            target,
            Children([NumberedListItemBlocktype(rich_text, block_color=block_color)]),
        )
        return numbered_list_block

    @staticmethod
    def to_do(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
        checked: Optional[bool] = False,
    ) -> Block:
        """https://developers.notion.com/reference/block#to-do-blocks"""
        to_do_block = BlockFactory._append(
            target,
            Children(
                [ToDoBlocktype(rich_text, block_color=block_color, checked=checked)]
            ),
        )
        return to_do_block

    @staticmethod
    def toggle(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#toggle-blocks"""

        toggle_block = BlockFactory._append(
            target, Children([ToggleBlocktype(rich_text, block_color=block_color)])
        )
        return toggle_block

    @staticmethod
    def code(
        target: Union[Page, Block, MutableMapping[str, Any]],
        rich_text: Sequence[Union[RichText, Mention, Equation]],
        /,
        *,
        language: Optional[Union[CodeBlockLang, str]] = None,
        caption: Optional[Sequence[Union[RichText, Mention, Equation]]] = None,
    ) -> Block:
        """
        :raises:
        ```py
        notion.exceptions.errors.NotionValidationError: body failed validation:
        body.children[0].code.rich_text[0].text.content.length should be ≤ `2000`, instead was `57839`.

        https://developers.notion.com/reference/block#code-blocks
        """
        code_block = BlockFactory._append(
            target,
            Children([CodeBlocktype(rich_text, language=language, caption=caption)]),
        )
        return code_block

    @staticmethod
    def embed_url(
        target: Union[Page, Block, MutableMapping[str, Any]], embedded_url: str, /
    ) -> Block:
        """
        ##### WARNING: embed block types through the API are not reliable and often cause errors.

        Embed block types are:
        Framer, Twitter (tweets), Google Drive documents, Gist, Figma,
        Invision, Loom, Typeform, Codepen, PDFs, Google Maps, Whimisical,
        Miro, Abstract, excalidraw, Sketch, Replit
        There is no need to specify the specific embed type, only the URL.

        ### Differences in embed blocks between the Notion app and the API
        The Notion app uses a 3rd-party service, Embedly, to validate and request metadata for embeds given a URL.
        This works well in a web app because Notion can kick off an asynchronous request for URL information,
        which might take seconds or longer to complete, and then update the block with the metadata in the UI
        after receiving a response from Embedly.

        Embedly is not called when creating embed blocks in the API because the API needs to be able to return
        faster than the UI, and because the response from Embedly could actually cause us change the block type.
        This would result in a slow and potentially confusing experience as the block in the response would
        not match the block sent in the request.
        The result is that embed blocks created via the API may not look exactly
        like their counterparts created in the Notion app.

        https://developers.notion.com/reference/block#embed-blocks
        """
        embed_url_block = BlockFactory._append(
            target, Children([EmbedBlocktype(embedded_url)])
        )
        return embed_url_block

    @staticmethod
    def bookmark(
        target: Union[Page, Block, MutableMapping[str, Any]],
        bookmark_url: str,
        /,
        *,
        caption: Optional[Sequence[Union[RichText, Mention, Equation]]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#bookmark-blocks"""

        bookmark_block = BlockFactory._append(
            target, Children([BookmarkBlocktype(bookmark_url, caption=caption)])
        )
        return bookmark_block

    @staticmethod
    def link_to_page(
        target: Union[Page, Block, MutableMapping[str, Any]], page_id: str, /
    ) -> Block:
        """https://developers.notion.com/reference/block#link-to-page-blocks"""

        link_to_page_block = BlockFactory._append(
            target, Children([LinkToPageBlockType(page_id)])
        )
        return link_to_page_block

    @staticmethod
    def equation(
        target: Union[Page, Block, MutableMapping[str, Any]], expression: str, /
    ) -> Block:
        """
        :param expression: (required) A KaTeX compatible string

        ---
        https://developers.notion.com/reference/block#equation-blocks
        """
        equation_block = BlockFactory._append(
            target, Children([EquationBlocktype(expression)])
        )
        return equation_block

    @staticmethod
    def table_of_contents(
        target: Union[Page, Block, MutableMapping[str, Any]],
        /,
        *,
        block_color: Optional[Union[BlockColor, str]] = None,
    ) -> Block:
        """https://developers.notion.com/reference/block#table-of-contents-blocks"""
        table_of_contents_block = BlockFactory._append(
            target, Children([TableOfContentsBlocktype(block_color=block_color)])
        )
        return table_of_contents_block


# TODO
# "FileBlocktype",
# "PdfBlocktype",
# ImageBlocktype,
# VideoBlocktype,
# "TemplateBlocktype",
# "ColumnBlocktype",
# "ColumnList",
# "TableBlocktype",
# "TableRowBlocktype"



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Optional, Sequence, Union

from notion.properties.build import NotionObject

__all__: Sequence[str] = ("SortFilter", "PropertyValueSort", "EntryTimestampSort")


class SortFilter(NotionObject):
    """
    A sort is a condition used to order the entries returned from a database query.
    A database query can be sorted by a property and/or timestamp and in a given direction.
    For example,
    a library database can be sorted by the "Name of a book" (i.e. property) and in ascending (i.e. direction).

    Database queries can also be sorted by two or more properties, which is formally called a nested sort.
    The sort object listed first in the nested sort list takes precedence.

    :param sort_object: (required) List containing one of
        `notion.query.sort.PropertyValueSort` or `notion.query.sort.EntryTimestampSort`

    https://developers.notion.com/reference/post-database-query-sort#sort-object"""

    __slots__: Sequence[str] = ()

    def __init__(
        self, sort_object: list[Union[PropertyValueSort, EntryTimestampSort]]
    ) -> None:
        super().__init__()
        self.set("sorts", sort_object)


class PropertyValueSort(NotionObject):
    """
    This sort orders the database query by a particular property.
    https://developers.notion.com/reference/post-database-query-sort#sort-object
    """

    __slots__: Sequence[str] = ()

    def __init__(
        self, property_name: str, /, *, direction: Optional[str] = None
    ) -> None:
        super().__init__()
        self.set("property", property_name)
        self.set("direction", direction)

    @classmethod
    def ascending(
        cls, property_name: str, /, *, direction: Optional[str] = "ascending"
    ) -> PropertyValueSort:
        return cls(property_name, direction=direction)

    @classmethod
    def descending(
        cls, property_name: str, /, *, direction: Optional[str] = "descending"
    ) -> PropertyValueSort:
        return cls(property_name, direction=direction)


class EntryTimestampSort(NotionObject):
    """
    This sort orders the database query by the timestamp associated with a database entry.

    Required:
    - must use either `created_time_ascending` or `created_time_descending` classmethod.
    - must use either `last_edited_time_ascending` or `last_edited_time_descending` classmethod.

    https://developers.notion.com/reference/post-database-query-sort#entry-timestamp-sort
    """

    __slots__: Sequence[str] = ("_timestamp", "_direction")

    def __init__(self) -> None:
        super().__init__()

        self._timestamp: str
        self._direction: str
        self.set("timestamp", self._timestamp)
        self.set("direction", self._direction)

    @classmethod
    def created_time_ascending(cls) -> EntryTimestampSort:
        cls._timestamp = "created_time"
        cls._direction = "ascending"
        return cls()

    @classmethod
    def created_time_descending(cls) -> EntryTimestampSort:
        cls._timestamp = "created_time"
        cls._direction = "descending"
        return cls()

    @classmethod
    def last_edited_time_ascending(cls) -> EntryTimestampSort:
        cls._timestamp = "last_edited_time"
        cls._direction = "ascending"
        return cls()

    @classmethod
    def last_edited_time_descending(cls) -> EntryTimestampSort:
        cls._timestamp = "last_edited_time"
        cls._direction = "descending"
        return cls()



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import logging
import os
from types import ModuleType
from typing import (
    Any,
    Iterable,
    MutableMapping,
    Optional,
    Sequence,
    TypeAlias,
    Union,
    cast,
)

try:
    import orjson

    default_json: ModuleType = orjson
except ModuleNotFoundError:
    import json

    default_json: ModuleType = json  # type: ignore[no-redef]

import requests

from notion.api._about import *
from notion.exceptions import *

__all__: Sequence[str] = ["_NotionClient"]

logging.basicConfig(level=logging.INFO)

_NLOG = logging.getLogger("notion-api")


class _NotionClient:
    """Base Class to inherit: token, headers, requests, and endpoints."""

    def __init__(
        self, *, token: Optional[str] = None, notion_version: Optional[str] = None
    ) -> None:
        if token:
            self.token = token
        else:
            try:
                self.token = os.environ["NOTION_TOKEN"]
            except NameError:
                pass
            finally:
                if self.token is None:
                    raise NotionUnauthorized(
                        (
                            f"notion.{self.__class__.__name__} Missing Token, ",
                            "Check .env config | if token is named `NOTION_TOKEN`",
                        )
                    )

        __auth__ = f"Bearer {self.token}"

        self.headers: dict[str, str] = {
            "Authorization": __auth__,
            "Accept": __content_type__,
            "Content-type": __content_type__,
            "Notion-Version": __notion_version__,
        }

        if notion_version is not None:
            self.headers["Notion-Version"] = notion_version

    NotionEndpoint: TypeAlias = str

    @staticmethod
    def _block_endpoint(
        object_id: Optional[str] = None,
        /,
        *,
        children: Optional[bool] = None,
        page_size: Optional[int] = None,
        start_cursor: Optional[str] = None,
    ) -> NotionEndpoint:
        object_id_ = f"/{object_id}" if object_id else ""
        children_ = "/children" if children else ""
        page_size_ = f"&page_size={page_size}" if page_size else ""
        start_cursor_ = f"&start_cursor={start_cursor}" if start_cursor else ""
        urlparam_ = ""
        if any([page_size, start_cursor]):
            urlparam_ = "?"

        return "{}blocks{}{}{}{}{}".format(
            __base_url__, object_id_, children_, urlparam_, start_cursor_, page_size_
        )

    @staticmethod
    def _database_endpoint(
        object_id: Optional[str] = None, /, *, query: Optional[bool] = False
    ) -> NotionEndpoint:
        object_id_ = f"/{object_id}" if object_id else ""
        query_ = "/query" if query else ""

        return f"{__base_url__}databases{object_id_}{query_}"

    @staticmethod
    def _pages_endpoint(
        object_id: Optional[str] = None,
        /,
        *,
        properties: Optional[bool] = False,
        property_id: Optional[str] = None,
    ) -> NotionEndpoint:
        object_id_ = f"/{object_id}" if object_id else ""
        properties_ = "/properties" if properties else ""
        property_id_ = f"/{property_id}" if property_id else ""

        return f"{__base_url__}pages{object_id_}{properties_}{property_id_}"

    def _get(
        self,
        url: NotionEndpoint,
        /,
        *,
        payload: Optional[
            Union[MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]]
        ] = None,
    ) -> MutableMapping[str, Any]:
        if payload is None:
            response = cast(
                "MutableMapping[str, Any]",
                default_json.loads(requests.get(url, headers=self.headers).text),
            )
        else:
            if isinstance(payload, dict):
                payload = default_json.dumps(payload)
            response = cast(
                "MutableMapping[str, Any]",
                default_json.loads(
                    requests.post(url, headers=self.headers, json=payload).text
                ),
            )

        validate_response(response)
        return response

    def _post(
        self,
        url: NotionEndpoint,
        /,
        *,
        payload: Optional[
            Union[MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]]
        ] = None,
    ) -> MutableMapping[str, Any]:
        if payload is None:
            response = cast(
                "MutableMapping[str, Any]",
                default_json.loads(requests.post(url, headers=self.headers).text),
            )
        else:
            if isinstance(payload, dict):
                payload = default_json.dumps(payload)
            response = cast(
                "MutableMapping[str, Any]",
                default_json.loads(
                    requests.post(url, headers=self.headers, data=payload).text
                ),
            )

        validate_response(response)
        return response

    def _patch(
        self,
        url: NotionEndpoint,
        /,
        *,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:  # -> MutableMapping[str, Any]:
        if isinstance(payload, dict):
            payload = default_json.dumps(payload)
        response = cast(
            "MutableMapping[str, Any]",
            default_json.loads(
                requests.patch(url, headers=self.headers, data=payload).text
            ),
        )

        validate_response(response)
        return response

    def _delete(self, url: NotionEndpoint, /) -> MutableMapping[str, Any]:
        response = cast(
            "MutableMapping[str, Any]",
            default_json.loads(requests.delete(url, headers=self.headers).text),
        )

        validate_response(response)
        return response



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Optional, Sequence

from notion.properties.build import NotionObject

__all__: Sequence[str] = (
    "Parent",
    "UserObject",
    "NotionUUID",
    "NotionURL",
)


class Parent(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, id: str, /, *, type: str) -> None:
        super().__init__()
        self.nest("parent", "type", type)
        self.nest("parent", type, id)

    @classmethod
    def page(cls, id: str, /) -> Parent:
        return cls(id, type="page_id")

    @classmethod
    def database(cls, id: str, /) -> Parent:
        return cls(id, type="database_id")

    @classmethod
    def block(cls, id: str, /) -> Parent:
        return cls(id, type="block_id")


class UserObject(NotionObject):
    """
    The User object represents a user in a Notion workspace. Users include full workspace members, and integrations.
    Guests are not included.

    ---
    :param id: (required) Unique identifier for this user. *Always present
    :param name: (optional) User's name, as displayed in Notion.
    :param avatar_url: (optional)
    :param email: (optional)

    ---
    User objects appear in the API in nearly all objects returned by the API, including:
        - Block object under created_by and last_edited_by.
        - Page object under created_by and last_edited_by and in people property items.
        - Database object under created_by and last_edited_by.
        - Rich text object, as user mentions.
        - Property object when the property is a people property.

    User objects will always contain object and id keys.
    The remaining properties may appear if the user is being rendered in a rich text or
    page property context, and the bot has the correct capabilities to access those properties.

    ---
    All parameters are display-only and cannot be updated in Notion.

    https://developers.notion.com/reference/user
    """

    __slots__: Sequence[str] = ()

    def __init__(
        self,
        id: str,
        name: Optional[str] = None,
        avatar_url: Optional[str] = None,
        email: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.set("object", "user")
        self.set("id", id)
        self.set("name", name) if name else None
        self.set("avatar_url", avatar_url) if avatar_url else None
        self.set("type", "person")
        if email:
            self.nest("person", "email", email)
        else:
            self.set("person", {})  # must at least be an empty object.


class NotionURL(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, url: str, /) -> None:
        super().__init__()
        self.set("url", url)


class NotionUUID(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, id: str, /) -> None:
        super().__init__()
        self.set("id", id)


# TODO
# Database Description
# Cursor



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The endpoint limits the page response object to 25 references across all page properties. 
If a page object's Parent object is a database, then the property values conform to the database property schema. 
If a page object is not part of a database, then the only property value available for that page is its title.
The Retrieve a page property item endpoint returns information for a single property.

2 types of responses: 
Retrieve a page endpoint response object (named_property), 
Retrieve a page property item endpoint response object (paginated list of results with values) 
---

For information about size limitations for specific page property objects, 
refer to https://developers.notion.com/reference/request-limits#limits-for-property-values

NOTE: The following property values cannot be updated directly via the API and are excluded.
- Created By
- Created Time
- Last Edited By
- Last Edited Time
- Formula property value objects represent the result of evaluating a formula described in the
    database's properties. They contain a formula object with a type of either boolean/date/number/string.
    The value can't be updated directly via the API.

FilesPropertyValue can be found in `notion.properties.files`.

https://developers.notion.com/reference/page-property-values
"""
from __future__ import annotations

import abc
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Sequence, Union

from notion.properties.build import NotionObject
from notion.properties.common import NotionUUID, UserObject
from notion.properties.options import FunctionFormat
from notion.properties.propertyobjects import Option
from notion.properties.richtext import Equation, Mention, RichText

if TYPE_CHECKING:
    from datetime import timedelta

__all__: Sequence[str] = (
    "RichTextPropertyValue",
    "TitlePropertyValue",
    "DatePropertyValue",
    "RelationPropertyValue",
    "StatusPropertyValue",
    "SelectPropertyValue",
    "MultiSelectPropertyValue",
    "CheckboxPropertyValue",
    "PeoplePropertyValue",
    "RollupPropertyValue",
    "EmailPropertyValue",
    "NumberPropertyValue",
    "PhoneNumberPropertyValue",
    "URLPropertyValue",
)


class PagePropertyValue(metaclass=abc.ABCMeta):
    def __init__(self, property_name: str) -> None:
        self.name = property_name


class RichTextPropertyValue(PagePropertyValue, NotionObject):
    r"""
    The purpose of the rich text property value is to provide the key `rich_text`, whereas the object
    `notion.properties.RichText` has the key `text`.
    ---
    (required)
    :param rich_text_array: An array of rich text objects.
    ---
    Using `shift+enter` for multi-line text blocks results in
    a separate `text` key with a newline escape. E.g.

    ```json
    },  // ... first text line above
    {
        "type": "text",
        "text": {
            "content": "\nthis is the second line of a block"
        }
    }
    ```
    ---
    If you hyperlink only part of a string in Notion, the string will be split and return
    as separate keys in the rich text object.
    Hyperlinking text to an internal Notion link will populate link/href
    with the UUID's following notion.so/{workspace name}/...

    ```json
    {
        "type": "text",
        "text": {
            "content": "hyperlink in text to notion link\n",
            "link": {
                "url": "/3a2ec1e9308b4fd5a5749a5ee5aeeff9?v=f19121cb8e6f4329aba62edef93c39dc&p=bc5d3abdf3a942e0b6a7d8a5c94b5dc9&pm=s"
            }          // database id // database view id // database page id
    },
    ```

    ---
    :The RichText Object: https://developers.notion.com/reference/rich-text
    :The RichText Property Value: https://developers.notion.com/reference/page-property-values#rich-text
    """
    __slots__: Sequence[str] = ["name"]

    def __init__(
        self,
        property_name: str,
        rich_text_array: Sequence[Union[RichText, Mention, Equation]],
        /,
    ) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "rich_text")
        self.set("rich_text", rich_text_array)


class TitlePropertyValue(PagePropertyValue, NotionObject):
    """
    :param title: (required) An array of rich text objects: `notion.properties.propertyvalues.RichText`

    ---
    https://developers.notion.com/reference/page-property-values#title
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, title_: Sequence[Union[RichText, Mention, Equation]], /) -> None:
        super().__init__(property_name="title")
        self.set_array(self.name, title_)


class DatePropertyValue(PagePropertyValue, NotionObject):
    """
    Notion uses ISO 8601 date and time for some endpoints, and YYYY/MM/DD for others.
    If a datetime object is passed to either parameter, they'll be converted to isoformat.

    ---
    :param start: (required) A date, with an optional time.
        If the "date" value is a range, then start represents the start of the range.
    :param end: (optional) A string representing the end of a date range.
        If the value is null, then the date value is not a range.

    https://developers.notion.com/reference/page-property-values#date
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(
        self,
        property_name: str,
        /,
        *,
        start: Union[str, datetime],
        end: Optional[Union[str, datetime]] = None,
    ) -> None:
        super().__init__(property_name=property_name)

        self.nest("date", "start", start)
        self.nest("date", "end", end) if end else None


class RelationPropertyValue(PagePropertyValue, NotionObject):
    """
    NOTE: updating a relation property value with an empty array will clear the list.

    ---
    :param related_ids: (required) An array of related page references.
        A page reference is an object with an id key and a string value (UUIDv4)
        corresponding to a page ID in another database.

    :param has_more: (optional) If a relation has more than 25 references, then the has_more value for
        the relation in the response object is true. If a relation doesn't
        exceed the limit, then has_more is false.

    https://developers.notion.com/reference/page-property-values#relation
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(
        self,
        property_name: str,
        *related_ids: list[NotionUUID],
        has_more: Optional[bool] = None,
    ) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "relation")
        self.set("relation", [x for x in related_ids][0])
        if has_more:
            self.set("has_more", True)


class StatusPropertyValue(PagePropertyValue, NotionObject):
    """
    :param status_option: (required) a single status option: `notion.properties.Option`
        containing `name` and `notion.properties.PropertyColor`

    https://developers.notion.com/reference/page-property-values#status
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, status_option: Option, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "status")
        self.set("status", status_option)


class SelectPropertyValue(PagePropertyValue, NotionObject):
    """
    When selecting options, If the select database property does not yet have an option by the input name,
    then the name will be added to the database schema if the integration also has write access to the parent database.
    NOTE: Commas (",") are not valid for select values.

    ---
    :param select_option: (required) a single select option: `notion.properties.Option`
        containing `name` and `notion.properties.PropertyColor`

    https://developers.notion.com/reference/page-property-values#select
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, select_option: Option, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "select")
        self.set("select", select_option)


class MultiSelectPropertyValue(PagePropertyValue, NotionObject):
    """
    The MultiSelectPropertyValue contains an array of `notion.properties.Option` objects.
    When selecting options, If the multi-select database property does not yet have an option by the input name,
    then the name will be added to the database schema if the integration also has write access to the parent database.
    NOTE: Commas (",") are not valid for select values.

    ---
    :param options_array: (required) array of `notion.properties.Option`
        containing `name` and `notion.properties.PropertyColor`

    https://developers.notion.com/reference/page-property-values#multi-select
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, options_array: list[Option], /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "multi_select")
        self.set("multi_select", options_array)


class CheckboxPropertyValue(PagePropertyValue, NotionObject):
    """
    :param checkbox_value: (required) Whether the checkbox is checked (true) or unchecked (false).

    https://developers.notion.com/reference/page-property-values#checkbox
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, checkbox_value: bool, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "checkbox")
        self.set("checkbox", checkbox_value)


class PeoplePropertyValue(PagePropertyValue, NotionObject):
    """
    :param user_array: (required) An array of user objects.
        The Retrieve a page endpoint can't be guaranteed to return more than 25 people
        per people page property. If a people page property includes more than 25 people,
        then you can use the Retrieve a page property item endpoint for the specific
        people property to get a complete list of people.

    https://developers.notion.com/reference/page-property-values#people
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, user_array: list[UserObject], /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "people")
        self.set("people", user_array)


class RollupPropertyValue(PagePropertyValue, NotionObject):
    """
    :param function: (required) `notion.properties.FunctionFormat`, function to update property to.

    ---
    If the results of the rollup is a date (e.g. function = latest date), the results
    will be a DatePropertyValue.
    If the results of the rollup is a number (e.g. function = sum), the results will be
    a NumberPropertyValue.
    If the results of the rollup is anything requireing multiple values (e.g. function = show original),
    the results will be an array object, with a key 'type' = 'array',
    and key 'array' containing a list of values.

    NOTE: Only the function key can be updated via the API.

    https://developers.notion.com/reference/page-property-values#rollup
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, function: FunctionFormat, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "rollup")
        self.nest("rollup", "function", function)


class EmailPropertyValue(PagePropertyValue, NotionObject):
    """https://developers.notion.com/reference/page-property-values#email"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, email: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "email")
        self.set("email", email)


class NumberPropertyValue(PagePropertyValue, NotionObject):
    """https://developers.notion.com/reference/page-property-values#number"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, number: Union[float, timedelta], /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "number")
        self.set("number", number)


class PhoneNumberPropertyValue(PagePropertyValue, NotionObject):
    """
    :param phone_number: (required) A string representing a phone number.
        No phone number format is enforced.

    https://developers.notion.com/reference/page-property-values#phone-number
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, phone_number: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "phone_number")
        self.set("phone_number", phone_number)


class URLPropertyValue(PagePropertyValue, NotionObject):
    """https://developers.notion.com/reference/page-property-values#url"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, url: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "url")
        self.set("url", url)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Metadata that controls how a database property behaves. 
Each database property object contain the following; 
 - The ID of the property, usually a short string of random letters and symbols.
   Some automatically generated property types have special human-readable IDs. 
   For example, all Title properties have an ID of "title".name, id, and  key corresponding 
 - The name of the property as it appears in Notion.
 - A key corresponding with the value of type. The value is an object containing 
   type-specific configuration that controls the behavior of the property

NOTE: It is not possible to update a status database property in the current version of the Notion API. 
      Update these values from the Notion UI, instead.

https://developers.notion.com/reference/property-object
"""
from __future__ import annotations

import abc
from typing import Optional, Sequence, Union

from notion.exceptions import errors
from notion.properties.build import NotionObject
from notion.properties.options import FunctionFormat, NumberFormats, PropertyColor

__all__: Sequence[str] = (
    "RelationPropertyObject",
    "MultiSelectPropertyObject",
    "SelectPropertyObject",
    "NumberPropertyObject",
    "FormulaPropertyObject",
    "CheckboxPropertyObject",
    "PeoplePropertyObject",
    "PhoneNumberPropertyObject",
    "RichTextPropertyObject",
    "CreatedTimePropertyObject",
    "CreatedByPropertyObject",
    "LastEditedTimePropertyObject",
    "LastEditedByPropertyObject",
    "DatePropertyObject",
    "EmailPropertyObject",
    "FilesPropertyObject",
    "TitlePropertyObject",
    "URLPropertyObject",
    "RollupPropertyObject",
    "Option",
)


class PropertyObject(metaclass=abc.ABCMeta):
    def __init__(self, property_name: str) -> None:
        self.name = property_name


class TitlePropertyObject(PropertyObject, NotionObject):
    """
    A title database property controls the title that appears at the top of a page when a
    database row is opened. The title type object itself is empty; there is no additional configuration.

    NOTE: All databases require one, and only one, title property.
          The API throws errors if you send a request to Create a database without a title property,
          or if you attempt to Update a database to add or remove a title property.

    ---
    ### Title database property vs. database title
    A title database property is a type of column in a database.
    A database title defines the title of the database and is found on the database object.
    Every database requires both a database title and a title database property.

    https://developers.notion.com/reference/property-object#title
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "title")
        self.set("title", {})


class _Dual_Property(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, database_id: str, synced_property_name: str) -> None:
        super().__init__()
        self.set("database_id", database_id)
        self.set("type", "dual_property")
        self.nest("dual_property", "synced_property_name", synced_property_name)


class _Single_Property(NotionObject):
    __slots__: Sequence[str] = ()

    def __init__(self, database_id: str, /) -> None:
        super().__init__()
        self.set("database_id", database_id)
        self.set("type", "single_property")
        self.set("single_property", {})


class RelationPropertyObject(PropertyObject, NotionObject):
    """
    Use either classmethod: `dual`/`single`

    https://developers.notion.com/reference/property-object#relation
    """

    __slots__: Sequence[str] = ("name", "_related_to_")

    def __init__(
        self,
        property_name: str,
        /,
        *,
        database_id: str,
        synced_property_name: Optional[str] = None,
    ) -> None:
        super().__init__(property_name=property_name)
        self._related_to_: Union[_Dual_Property, _Single_Property]

        try:
            self.set("type", "relation")
            self.set("relation", self._related_to_)
        except AttributeError:
            raise errors.NotionInvalidJson("Use classmethods.")

    @classmethod
    def dual(
        cls, property_name: str, database_id: str, synced_property_name: str, /
    ) -> RelationPropertyObject:
        """
        :param database_id: (required) The database that the relation property refers to.
            The corresponding linked page values must belong to the database in order to be valid.
        :param synced_property_name: (required) The name of the corresponding property that is
            updated in the related database when this property is changed.
        """
        cls._related_to_ = _Dual_Property(database_id, synced_property_name)
        return cls(
            property_name,
            database_id=database_id,
            synced_property_name=synced_property_name,
        )

    @classmethod
    def single(cls, property_name: str, database_id: str, /) -> RelationPropertyObject:
        """
        :param database_id: (required) The database that the relation property refers to.
            The corresponding linked page values must belong to the database in order to be valid.
        """
        cls._related_to_ = _Single_Property(database_id)
        return cls(property_name, database_id=database_id)


class Option(NotionObject):
    """
    :param name: (required) The name of the option as it appears in the Notion UI.
        Note: Commas (",") are not valid for select values.
    :param color: (required) The color of the option as rendered in the Notion UI.
        Use `notion.properties.PropertyColor` for reference.
    """

    __slots__: Sequence[str] = ()

    def __init__(
        self, option_name: str, color: Optional[Union[PropertyColor, str]] = None, /
    ) -> None:
        super().__init__()
        self.set("name", option_name)
        if not color:
            self.set("color", PropertyColor.default)
        else:
            self.set("color", color)


class MultiSelectPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#multi-select"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /, *, options: list[Option]) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "multi_select")
        self.nest("multi_select", "options", options)


class SelectPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#select"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /, *, options: list[Option]) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "select")
        self.nest("select", "options", options)


class NumberPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#number"""

    __slots__: Sequence[str] = ["name"]

    def __init__(
        self,
        property_name: str,
        format: Optional[Union[NumberFormats, str]] = None,
        /,
    ) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "number")
        if not format:
            self.nest("number", "format", NumberFormats.number)
        else:
            self.nest("number", "format", format)


class FormulaPropertyObject(PropertyObject, NotionObject):
    """
    :param expression: (required) The formula that is used to compute the values for this property.
        Refer to the Notion help center for information about formula syntax.
    https://developers.notion.com/reference/property-object#formula
    """

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, expression: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "formula")
        self.nest("formula", "expression", expression)


class CheckboxPropertyObject(PropertyObject, NotionObject):
    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "checkbox")
        self.set("checkbox", {})


class PeoplePropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#people"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "people")
        self.set("people", {})


class PhoneNumberPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#phone-number"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "phone_number")
        self.set("phone_number", {})


class RichTextPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#rich-text"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "rich_text")
        self.set("rich_text", {})


class CreatedTimePropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#created-time"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "created_time")
        self.set("created_time", {})


class CreatedByPropertyObject(PropertyObject, NotionObject):
    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "created_by")
        self.set("created_by", {})


class LastEditedTimePropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#last-edited-time"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "last_edited_time")
        self.set("last_edited_time", {})


class LastEditedByPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#last-edited-by"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "last_edited_by")
        self.set("last_edited_by", {})


class DatePropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#date"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "date")
        self.set("date", {})


class EmailPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#email"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "email")
        self.set("email", {})


class FilesPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#files"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "files")
        self.set("files", {})


class URLPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#url"""

    __slots__: Sequence[str] = ["name"]

    def __init__(self, property_name: str, /) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "url")
        self.set("url", {})


class RollupPropertyObject(PropertyObject, NotionObject):
    """https://developers.notion.com/reference/property-object#rollup"""

    __slots__: Sequence[str] = ["name"]

    def __init__(
        self,
        property_name: str,
        relation_property_name: str,
        rollup_property_name: str,
        function: Union[FunctionFormat, str],
        /,
    ) -> None:
        super().__init__(property_name=property_name)
        self.set("type", "rollup")
        self.nest("rollup", "relation_property_name", relation_property_name)
        self.nest("rollup", "rollup_property_name", rollup_property_name)
        self.nest("rollup", "function", function)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from functools import cached_property
from typing import Any, Iterable, MutableMapping, Optional, Sequence, Union

from notion.api.blockmixin import _TokenBlockMixin
from notion.api.client import _NLOG

__all__: Sequence[str] = ["Block"]


class Block(_TokenBlockMixin):
    """A block object represents content within Notion.
    Blocks can be text, lists, media, and more. A page is a type of block, too.

    These are the individual 'nodes' in a page that you typically interact with in Notion.
    Some blocks have more content nested inside them.
    Some examples are indented paragraphs, lists, and toggles.
    The nested content is called children, and children are blocks, too.

    ---
    :param id: (required) `block_id` of object in Notion.
    :param token: (required) Bearer token provided when you create an integration.
        set as `NOTION_TOKEN` in .env or set variable here.
        see https://developers.notion.com/reference/authentication.
    :param notion_version: (optional) API version
        see https://developers.notion.com/reference/versioning

    https://developers.notion.com/reference/block
    """

    def __init__(
        self,
        id: str,
        /,
        *,
        token: Optional[str] = None,
        notion_version: Optional[str] = None,
    ) -> None:
        super().__init__(id, token=token, notion_version=notion_version)

        self.logger = _NLOG.getChild(f"{self.__repr__()}")

    @cached_property
    def retrieve(self) -> MutableMapping[str, Any]:
        """
        Retrieves a Block object using the ID specified.

        https://developers.notion.com/reference/retrieve-a-block
        """
        return self._get(self._block_endpoint(self.id))

    def retrieve_children(
        self, start_cursor: Optional[str] = None, page_size: Optional[int] = None
    ) -> MutableMapping[str, Any]:
        """
        Returns a paginated array of child block objects contained
        in the block using the ID specified.

        Returns only the first level of children for the specified block.
        See block objects for more detail on determining if that block has nested children.
        In order to receive a complete representation of a block,
        you may need to recursively retrieve block children of child blocks.
        page_size Default: 100 page_size Maximum: 100.

        https://developers.notion.com/reference/get-block-children
        """
        return self._get(
            self._block_endpoint(
                self.id, children=True, page_size=page_size, start_cursor=start_cursor
            )
        )

    def _append(
        self,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:
        """
        Creates/appends new children blocks to the parent block_id specified.
        Returns a paginated list of newly created children block objects.

        Used internally by `notion.api.blocktypefactory.BlockFactory`.

        https://developers.notion.com/reference/patch-block-children
        """
        return self._patch(
            self._block_endpoint(self.id, children=True), payload=payload
        )

    @property
    def delete_self(self) -> None:
        """
        Sets a Block object, including page blocks, to archived: true
        using the ID specified. Note: in the Notion UI application,
        this moves the block to the "Trash" where it can still be
        accessed and restored. To restore the block with the API,
        use the Update a block or Update page respectively.

        https://developers.notion.com/reference/delete-a-block
        """
        self._delete(self._block_endpoint(self.id))
        self.logger.info("Deleted Self.")

    @property
    def restore_self(self) -> None:
        """
        Sets "archived" key to false.
        Only works if the parent page has not been deleted from the trash.
        """
        self._patch(self._block_endpoint(self.id), payload=(b'{"archived": false}'))
        self.logger.info("Restored Self.")

    def delete_child(self, children_id: list[str]) -> None:
        for id in children_id:
            self._delete(self._block_endpoint(self.id))
            self.logger.info(f"Deleted child block `{id}`.")

    def restore_child(self, children_id: list[str]) -> None:
        for id in children_id:
            self._patch(self._block_endpoint(self.id), payload=b'{"archived": false}')
            self.logger.info(f"Restored child block `{id}`.")

    def update(
        self,
        payload: Union[
            MutableMapping[str, Any], Union[Iterable[bytes], bytes, bytearray]
        ],
    ) -> MutableMapping[str, Any]:
        """
        Updates content for the specified block_id based on the block
        type. Supported fields based on the block object type.
        Note: The update replaces the entire value for a given field.
        If a field is omitted (ex: omitting checked when updating a to_do block),
        the value will not be changed.

        To update title of a child_page block, use the pages endpoint
        To update title of a child_database block, use the databases endpoint

        Toggle can be added and removed from a heading block. However,
        you cannot remove toggle from a heading block if it has children
        All children MUST be removed before revoking toggle from a heading block

        https://developers.notion.com/reference/update-a-block
        """
        return self._patch(self._block_endpoint(self.id), payload=payload)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Type-specific filter conditions
https://developers.notion.com/reference/post-database-query-filter#type-specific-filter-conditionss
"""
from typing import Literal, Sequence, TypeAlias, Union

__all__: Sequence[str] = (
    "TextConditions",
    "TextTypes",
    "NumberConditions",
    "CheckboxConditions",
    "SelectConditions",
    "MultiSelectConditions",
    "StatusConditions",
    "DateConditions",
    "DateTypes",
    "PeopleConditions",
    "PeopleTypes",
    "FilesConditions",
    "RelationConditions",
    "RollupConditions",
    "FormulaConditions",
    "FilterConditions",
)


TextConditions: TypeAlias = Literal[
    "equals",
    "does_not_equal",
    "contains",
    "does_not_contain",
    "starts_with",
    "ends_with",
    "is_empty",
    "is_not_empty",
]

TextTypes: TypeAlias = Literal[
    "title",
    "rich_text",
    "url",
    "email",
]

NumberConditions: TypeAlias = Literal[
    "equals",
    "does_not_equal",
    "greater_than",
    "less_than",
    "greater_than_or_equal_to",
    "less_than_or_equal_to",
    "is_empty",
    "is_not_empty",
]

CheckboxConditions: TypeAlias = Literal["equals", "does_not_equal"]

SelectConditions: TypeAlias = Literal[
    "equals", "does_not_equal", "is_empty", "is_not_empty"
]

MultiSelectConditions: TypeAlias = Literal[
    "contains", "does_not_contain", "is_empty", "is_not_empty"
]

StatusConditions: TypeAlias = Literal[
    "equals", "does_not_equal", "is_empty", "is_not_empty"
]

DateConditions: TypeAlias = Literal[
    "equals",
    "before",
    "after",
    "on_or_before",
    "is_empty",
    "is_not_empty",
    "on_or_after",
    "past_week",
    "past_month",
    "past_year",
    "this_week",
    "next_week",
    "next_month",
    "next_year",
]

DateTypes: TypeAlias = Literal[
    "date",
    "created_time",
    "last_edited_time",
]

PeopleConditions: TypeAlias = Literal[
    "contains", "does_not_contain", "is_empty", "is_not_empty"
]

PeopleTypes: TypeAlias = Literal[
    "people",
    "created_by",
    "last_edited_by",
]

FilesConditions: TypeAlias = Literal["is_empty", "is_not_empty"]

RelationConditions: TypeAlias = Literal[
    "contains", "does_not_contain", "is_empty", "is_not_empty"
]

RollupConditions: TypeAlias = Literal["any", "every", "none", "number", "date"]

FormulaConditions = Union[
    TextConditions, NumberConditions, CheckboxConditions, DateConditions
]

FilterConditions: TypeAlias = Union[
    TextConditions,
    TextTypes,
    NumberConditions,
    CheckboxConditions,
    SelectConditions,
    MultiSelectConditions,
    StatusConditions,
    DateConditions,
    DateTypes,
    PeopleConditions,
    PeopleTypes,
    FilesConditions,
    RelationConditions,
    RollupConditions,
]



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from types import ModuleType
from typing import Any, Iterable, MutableMapping, Optional, Sequence, Union

try:
    import orjson

    default_json: ModuleType = orjson
except ModuleNotFoundError:
    import json

    default_json: ModuleType = json  # type: ignore[no-redef]

__all__: Sequence[str] = ["build_payload"]


def build_payload(*objects: dict[str, Any]) -> Union[Iterable[bytes], bytes, bytearray]:
    final: dict[str, Any] = {}
    for o in objects:
        final |= o
    return default_json.dumps(final)  # type: ignore[no-any-return]


class NotionObject(dict[str, Any]):
    __slots__: Sequence[str] = ()

    def __init__(self) -> None:
        super().__init__()

    def set(self, k: str, v: Any) -> None:
        self[k] = v

    def nest(self, key: str, k: Optional[str], v: Any) -> None:
        if key not in self:
            self.set(key, {k: v})
        else:
            self[key] |= {k: v}

    def set_array(
        self,
        key: str,
        values: Union[
            Iterable[bytes],
            bytes,
            bytearray,
            MutableMapping[str, Any],
            Sequence[Any],
        ],
    ) -> None:
        self[key] = list(values)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" 
Visit https://developers.notion.com/reference/post-database-query for info on database queries and filter/sort objects.

`from notion.query import *`

`notion.query.PropertyFilter` && `notion.query.TimestampFilter` for individual filters as used in Notion UI.

`notion.query.CompoundFilter` to combine filters.
    :method _and(): combine all filters in an `and` grouping.
    :method _or(): combine all filters in an `or` grouping.

    Create a separate CompoundFilter object to nest an `and` operator inside another `and` or `or`.

`notion.query.SortFilter` contains a list of either *`notion.query.PropertyValueSort` | `notion.query.EntryTimestampSort`
"""
from typing import Sequence

from notion.query.compound import *
from notion.query.conditions import *
from notion.query.propfilter import *
from notion.query.sort import *
from notion.query.timestamp import *

__all__: Sequence[str] = (
    "SortFilter",
    "PropertyFilter",
    "CompoundFilter",
    "TimestampFilter",
    "EntryTimestampSort",
    "PropertyValueSort",
)



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any, MutableMapping, Optional, Sequence, Union

from notion.properties.build import NotionObject
from notion.query.conditions import DateConditions

__all__: Sequence[str] = ["TimestampFilter"]


class TimestampFilter(NotionObject):
    """
    A timestamp filter object must contain a timestamp key corresponding to the type of timestamp
    and a key matching that timestamp type which contains a date filter condition.
    Possible values are: "created_time", "last_edited_time".

    Required: must use either `created_time` or `last_edited_time` classmethod.

    :param filter_condition: One of the available DateConditions. Must be a string (ISO 8601 date).
                          for the following date conditions, the filter_value must be '{}'
                          [past_week, past_month, past_year, this_week, next_week, next_month, next_year]

    :param filter_value: Value to filter for.
                         If a date is provided, the comparison is done against the start and end of the UTC date.
                         If a date with a time is provided, the comparison is done with millisecond precision.
                         Note that if no timezone is provided, the default is UTC.

    https://developers.notion.com/reference/post-database-query-filter#timestamp-filter-object
    """

    __slots__: Sequence[str] = ("_filter_condition", "_filter_type", "_type")

    def __init__(
        self,
        filter_condition: DateConditions,
        filter_value: Union[str, MutableMapping[str, Any]],
        /,
        *,
        _type: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.nest("filter", "timestamp", _type)
        self.nest("filter", _type, {filter_condition: filter_value})

    @classmethod
    def created_time(
        cls,
        filter_condition: DateConditions,
        filter_value: Union[str, MutableMapping[str, Any]],
        /,
        *,
        _type: Optional[str] = None,
    ) -> TimestampFilter:
        return cls(filter_condition, filter_value, _type="created_time")

    @classmethod
    def last_edited_time(
        cls,
        filter_condition: DateConditions,
        filter_value: Union[str, MutableMapping[str, Any]],
        /,
        *,
        _type: Optional[str] = None,
    ) -> TimestampFilter:
        return cls(filter_condition, filter_value, _type="last_edited_time")



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from datetime import datetime, tzinfo
from functools import cached_property
from typing import Any, MutableMapping, Optional, Sequence, Union
from uuid import UUID

from pytz import BaseTzInfo, timezone
from tzlocal import get_localzone

from notion.api._about import *
from notion.api.client import _NotionClient
from notion.exceptions.errors import NotionObjectNotFound

__all__: Sequence[str] = ["_TokenBlockMixin"]


class _TokenBlockMixin(_NotionClient):
    """
    Any object you interact with in Notion;
    Databases/Pages/individual child blocks, are all considered 'Blocks'
    This class assigns common attributes among all three types.
    """

    def __init__(
        self,
        id: str,
        /,
        *,
        token: Optional[str] = None,
        notion_version: Optional[str] = None,
    ) -> None:
        super().__init__(token=token, notion_version=notion_version)

        self.tz: BaseTzInfo = get_localzone()
        self.id: str = id.replace("-", "")

        try:
            UUID(self.id)
        except ValueError:
            raise NotionObjectNotFound(
                (
                    f"{self.__repr__()} instatiation failed validation: ",
                    f"id should be a valid uuid, instead was `'{self.id}'`",
                )
            )

    @cached_property
    def _block(self) -> MutableMapping[str, Any]:
        """
        Same result as retrieve() for `notion.api.notionblock.Block`.
        If used with `notion.api.notionpage.Page` or `notion.api.notiondatabase.Database`,
        retrieves the page or database object from the blocks endpoint.
        """
        return self._get(self._block_endpoint(self.id))

    @property
    def type(self) -> str:
        return str(self._block["type"])

    @property
    def object(self) -> str:
        return str(self._block["object"])

    @property
    def has_children(self) -> bool:
        return bool(self._block["has_children"])

    @property
    def is_archived(self) -> bool:
        return bool(self._block["archived"])

    @property
    def parent_type(self) -> str:
        return str(self._block["parent"]["type"])

    @property
    def parent_id(self) -> str:
        _parent_id = self._block["parent"][self.parent_type]
        if _parent_id is True:
            # sets key to 'workspace' rather than 'True'
            _parent_id = self.parent_type
        else:
            _parent_id = _parent_id.replace("-", "")
        return str(_parent_id)

    def set_tz(self, tz: Union[tzinfo, str]) -> None:
        """
        :param timezone: (required) set default timezone.
            class default matches the Windows-configured timezone
            Use `pytz.all_timezones` to retrieve list of tz options.
            Pass either str or `pytz.timezone(...)`
        """
        if isinstance(tz, tzinfo):
            self.__setattr__("tz", tz)
        elif isinstance(tz, str):
            self.__setattr__("tz", timezone(tz))

    @property
    def last_edited(self) -> datetime:
        """
        Notion returns datetime ISO 8601, UTC.
        Class default matches the Windows-configured timezone.
        Change default timezone with method `set_tz(...)`
        """
        return datetime.fromisoformat(self._block["last_edited_time"]).astimezone(
            tz=self.tz
        )

    @property
    def created_time(self) -> datetime:
        """
        Notion returns datetime ISO 8601, UTC.
        Class default matches the Windows-configured timezone.
        Change default timezone with method `set_tz(...)`
        """
        return datetime.fromisoformat(self._block["created_time"]).astimezone(
            tz=self.tz
        )

    def __repr__(self) -> str:
        return f"notion.{self.__class__.__name__}('{self.id}')"



# MIT License

# Copyright (c) 2023 ayvi-0001

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Sequence

__all__: Sequence[str] = (
    "_NotionErrors",
    "NotionInvalidJson",
    "NotionInvalidRequestUrl",
    "NotionInvalidRequest",
    "NotionValidationError",
    "NotionMissingVersion",
    "NotionUnauthorized",
    "NotionRestrictedResource",
    "NotionObjectNotFound",
    "NotionConflictError",
    "NotionRateLimited",
    "NotionInternalServerError",
    "NotionServiceUnavailable",
    "NotionDatabaseConnectionUnavailable",
)


class _NotionErrors(BaseException):
    """Base for an error raised by this API. Any exceptions should derive from this.

    More info on Errors and Request Limits at:
     - https://developers.notion.com/reference/errors
     - https://developers.notion.com/reference/request-limits
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NotionInvalidJson(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = ["Error 400: The request body could not be decoded as JSON."]


class NotionInvalidRequestUrl(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = ["Error 400: The request URL is not valid."]


class NotionInvalidRequest(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = ["Error 400: This request is not supported."]


class NotionValidationError(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 400: The request body does not match the schema for the expected parameters."
        ]


class NotionMissingVersion(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 400: The request is missing the required Notion-Version header. See Versioning."
        ]


class NotionUnauthorized(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = ["Error 401: The bearer token is not valid."]


class NotionRestrictedResource(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 403: Given the bearer token used, the client doesn't have permission to perform this operation."
        ]


class NotionObjectNotFound(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 404: Given the bearer token used, the resource does not exist. This error can also indicate that the resource has not been shared with owner of the bearer token."
        ]


class NotionConflictError(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 409: The transaction could not be completed, potentially due to a data collision. Make sure the parameters are up to date and try again."
        ]


class NotionRateLimited(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 429: This request exceeds the number of requests allowed. Slow down and try again. More details on rate limits."
        ]


class NotionInternalServerError(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 500: An unexpected error occurred. Reach out to Notion support."
        ]


class NotionServiceUnavailable(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 503: Notion is unavailable. Try again later. This can occur when the time to respond to a request takes longer than 60 seconds, the maximum request timeout."
        ]


class NotionDatabaseConnectionUnavailable(_NotionErrors):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

        self.__notes__ = [
            "Error 503: Notion's database is unavailable or in an unqueryable state. Try again later."
        ]
